------------------------------------------------------------------------------
--                L  O  C  A  L     V  A  R  I  A  B  L  E  S
-------------------------------------------------------------------------------

local addonName, L = ...;

local nglAchievementsFrame = nil;
local nglAchievementsEdge  = nil;
local nglAchievementsMsg   = nil;

local tabInfo =
{
  { -- Character
    filtered  = {string = "", categories = {}, achievements = {}},
    selected  =
    {
      string       = "",
      categoryId   = -1,
      filterValue  = 1,
      achievements = {},
    },
    history   = NigglesHistoryClass:New(),
  },
  { -- Guild
    filtered  = {string = "", categories = {}, achievements = {}},
    selected  =
    {
      string       = "",
      categoryId   = -1,
      filterValue  = 1,
      achievements = {},
    },
    history   = NigglesHistoryClass:New(),
  },
  { -- Stats
    filtered  = {string = "", categories = {}, achievements = {}},
    selected  =
    {
      string       = "",
      categoryId   = -1,
      filterValue  = 1,
      achievements = {},
    },
    history   = NigglesHistoryClass:New(),
  },
};

local currentTabInfo = tabInfo[1];

local blizzardAchievementsUpdate = nil;
local blizzardStatisticsUpdate   = nil;

local metaBtnHooked = {};

local itemRefAchievement =
{
  overlay = nil,
  id      = 0,
  link    = nil,
  text    = nil,
};

-------------------------------------------------------------------------------
--             L  O  C  A  L     D  E  F  I  N  I  T  I  O  N  S
-------------------------------------------------------------------------------

local achievementsFilterClear;
local achievementsHistoryBtnEnable;
local metaCriteriaPreClick;
local selectedCategoryUpdate;

-------------------------------------------------------------------------------
--                L  O  C  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to process the clicking of an Achievement Alert frame
--
local function achievementAlertOnClick(self, button, down)
  -- Local Variables
  local filterString;

  -- Check if the search functionality is enabled
  if ((NigglesDBC.settings.UiAchievementSearch == true) and
      (nglAchievementsFrame ~= nil)) then
    -- Check if there is any filtering
    filterString = nglAchievementsFrame.searchBox:GetText();
    if ((filterString ~= nil) and (filterString ~= "")) then
      -- Clear any filtering
      achievementsFilterClear(false);

      -- Call the standard function again, now the achievements are unfiltered
      AchievementAlertFrame_OnClick(self);
    end
  end

  return;
end

--
-- Function hooked into 'AchievementFrameCategories_Update()'. It updates
-- the font to indicate which categories have matches to any search string.
--
local function achievementCategoriesUpdate()
  -- Local Variables
  local buttons = AchievementFrameCategoriesContainer.buttons;
  local current;
  local filteredInfo = currentTabInfo.filtered;
  local numButtons

  -- Check if the selected category has changed
  if ((currentTabInfo.selected.categoryId ~=
        currentTabInfo.blizzard.selectedCategory) or
      (currentTabInfo.selected.filterValue ~=
        AchievementFrameFilterDropDown.value) or
      (currentTabInfo.selected.string ~=
        currentTabInfo.filtered.string)) then
    -- Update the achievements for the selected category
    selectedCategoryUpdate();
  end

  -- Update all the buttons
  numButtons = #buttons;
  for buttonIdx = 1, numButtons do
    -- Check if the current button is shown
    current = buttons[buttonIdx];
    if (current:IsShown()) then
      -- Check if there is no filtering or if the category has matches
      if ((filteredInfo.string == "") or
          (bit.band((filteredInfo.categories[current.categoryID] or 0),
            (4-AchievementFrameFilterDropDown.value)) ~= 0)) then
        if (type(current.element.parent) == "number") then
          current.label:SetFontObject("GameFontHighlight");
        else
          current.label:SetFontObject("GameFontNormal");
        end
      else
        current.label:SetFontObject("GameFontDisable");
      end
    end
  end

  return;
end

--
-- Function call when the History Back or Forward buttons are clicked.
--
local function achievementHistoryBtnOnClick(self, mouseButton)
  -- Local Variables
  local achievementId = 0;

  -- Check the feature is enabled
  if (NigglesDBC.settings.UiAchievementSearch) then
    -- Get the ID of of the achievement in the appropriate direction
    achievementId = currentTabInfo.history:Move(self:GetID() == 1);

    -- Check there was an achievement Id returned
    if (achievementId > 0) then
      -- Clear any filtering
      achievementsFilterClear(false);

      -- Select the achievement
      AchievementFrame_SelectAchievement(achievementId, true, false);
    end

    -- Enable/disable the history buttons
    achievementsHistoryBtnEnable();
  end

  return;
end

--
-- Function to hook the 'PreClick' script of new meta criteria buttons
--
local function achievementMetaBtnCreated(index)
  -- Check if the meta criteria button is already hooked
  if (not metaBtnHooked[index]) then
    -- Hook the button's 'PreClick' script
    _G["AchievementFrameMeta"..index]:HookScript("PreClick",
      metaCriteriaPreClick);
    metaBtnHooked[index] = true;
  end

  return;
end

--
-- Function called when the 'Completed' filter is changed
--
local function achievementsFilterOnChange()
  -- update the categories and achievements
  AchievementFrameCategories_Update();

  return;
end

--
-- Function hooked into "AchievementFrame_SetFilter". It updates the info for
-- the selected category.
--
local function achievementsFilterOnSet(value)
  -- Update the categories and achievements
  achievementCategoriesUpdate();

  return;
end

--
-- Function to update the Achievements
--
-- NOTE: This is a modified version of 'AchievementFrameAchievements_Update()'
--       in 'Blizzard_AchievementUI.lua', patch 7.0.3.
--
local function achievementsListUpdate()
  if (currentTabInfo.filtered.string == "") then
    blizzardAchievementsUpdate();
    return;
  end
  local category = currentTabInfo.blizzard.selectedCategory;
  if ( category == "summary" ) then
    return;
  end
  local scrollFrame = AchievementFrameAchievementsContainer

  local offset = HybridScrollFrame_GetOffset(scrollFrame);
  local buttons = scrollFrame.buttons;
  local numButtons = #buttons;

  local selectedAchiev = currentTabInfo.selected.achievements;
  local numAchievements = #selectedAchiev;

  -- If the current category is feats of strength and there are no entries then show the explanation text
  if ( AchievementFrame_IsFeatOfStrength() and numAchievements == 0 ) then
    if ( (AchievementFrame.selectedTab or 1) == 1 ) then
      AchievementFrameAchievementsFeatOfStrengthText:SetText(FEAT_OF_STRENGTH_DESCRIPTION);
    else
      AchievementFrameAchievementsFeatOfStrengthText:SetText(GUILD_FEAT_OF_STRENGTH_DESCRIPTION);
    end
    AchievementFrameAchievementsFeatOfStrengthText:Show();
  elseif (numAchievements == 0) then
    AchievementFrameAchievementsFeatOfStrengthText:SetText(L["No Matches"]);
    AchievementFrameAchievementsFeatOfStrengthText:Show();
  else
    AchievementFrameAchievementsFeatOfStrengthText:Hide();
  end

  local selection = AchievementFrameAchievements.selection;
  if ( selection ) then
    AchievementButton_ResetObjectives();
  end

  local extraHeight = scrollFrame.largeButtonHeight or ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT

  local achievementIndex;
  local displayedHeight = 0;
  for i = 1, numButtons do
    achievementIndex = i + offset;
    if (achievementIndex > numAchievements) then
      buttons[i]:Hide();
    else
      AchievementButton_DisplayAchievement(buttons[i], category,
        selectedAchiev[achievementIndex], selection);
      displayedHeight = displayedHeight + buttons[i]:GetHeight();
    end
  end

  local totalHeight = numAchievements * ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT;
  totalHeight = totalHeight + (extraHeight - ACHIEVEMENTBUTTON_COLLAPSEDHEIGHT);

  HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);

  if ( selection ) then
    AchievementFrameAchievements.selection = selection;
  else
    HybridScrollFrame_CollapseButton(scrollFrame);
  end
end

--
-- Function to update the Achievement Statistics
--
-- NOTE: This is a modified version of 'AchievementFrameStats_Update()'
--       in 'Blizzard_AchievementUI.lua', patch 7.0.3.
--
local function achievementsStatsUpdate()
  if (currentTabInfo.filtered.string == "") then
    blizzardStatisticsUpdate();
    return;
  end
  local category = currentTabInfo.blizzard.selectedCategory;
  local scrollFrame = AchievementFrameStatsContainer;
  local offset = HybridScrollFrame_GetOffset(scrollFrame);
  local buttons = scrollFrame.buttons;
  local numButtons = #buttons;
  local statHeight = 24;

  -- iterate through the displayStatCategories and display them
  local selectedAchiev = currentTabInfo.selected.achievements;
  local statCount = #selectedAchiev;
  local statIndex, id, button;
  local stat;

  -- Show/Hide the 'No Matches' message
  nglAchievementsMsg:SetShown(statCount == 0);

  local totalHeight = statCount * statHeight;
  local displayedHeight = numButtons * statHeight;
  for i = 1, numButtons do
    button = buttons[i];
    statIndex = offset + i;
    if ( statIndex <= statCount ) then
      stat = selectedAchiev[statIndex];
      if (stat < 0) then
        AchievementFrameStats_SetHeader(button, -stat);
      else
        AchievementFrameStats_SetStat(button, stat, nil, statIndex)
      end
      button:Show();
    else
      button:Hide();
    end
  end
  HybridScrollFrame_Update(scrollFrame, totalHeight, displayedHeight);
end

--
-- Function called when the Niggles Achievements search frame is shown
--
local function achievementSearchOnShow(self)
  -- Get the info for the current tab
  currentTabInfo = tabInfo[(AchievementFrame.selectedTab or 1)];

  -- Enable/disable the history buttons
  achievementsHistoryBtnEnable();

  return;
end

--
-- Function hooked into Blizzard functions used to display achievements. It
-- adds the achievements to the history for the relevant tab.
--
local function achievementOnSelect(buttonId, forceSelect, ignore)
  -- Local Variables
  local achievementId;

  -- Check if the selection should be ignored
  if ((NigglesDBC.settings.UiAchievementSearch) and (ignore ~= 1)) then
    -- Get the ID of the achievement selected
    achievementId = ((type(buttonId) == "table") and buttonId.id or buttonId);
    if (type(achievementId) == "number") then
      -- Add the achievement to the history
      currentTabInfo.history:Add(achievementId);
      achievementsHistoryBtnEnable();
    end
  end

  return;
end

--
-- Function to process events for the Achievement search box
--
--
local function achievementSearchOnEvent(self, event, ...)
  -- Process the event
  if (event == "ACHIEVEMENT_SEARCH_UPDATED") then
    -- Check there is a search string
    -- NOTE: This check is needed because there is a bug in the Blizzard API
    --       that means the search string can't be cleared and events are
    --       triggered when the achievement panel is opened.
    if (nglAchievementsFrame.searchBox:GetNumLetters() >=
         MIN_CHARACTER_SEARCH) then
      -- Local Variables
      local _;
      local achievements = currentTabInfo.filtered.achievements;
      local achievementId;
      local band = bit.band;
      local bor = bit.bor;
      local categories = currentTabInfo.filtered.categories;
      local categoryId;
      local parentId;
      local isCompleted;
      local completedFlag;

      -- Get the number of filtered achievements
      numFiltered = GetNumFilteredAchievements();

      -- Process the filtered achievements
      wipe(achievements);
      wipe(categories);
      for idx = 1, numFiltered do
        -- Get the ID of the achievement
        achievementId = GetFilteredAchievementID(idx);

        -- Check if the achievement is complete
        _, _, _, isCompleted = GetAchievementInfo(achievementId);

        -- Save the achievement ID
        achievements[achievementId] = (isCompleted and 0x02 or 0x01);

        -- Get the achievement's category
        categoryId = GetAchievementCategory(achievementId);
        if (categoryId ~= nil) then
          completedFlag = (isCompleted and 0x02 or 0x01);
          if (band((categories[categoryId] or 0), completedFlag) == 0) then
            -- Update the category
            categories[categoryId] =
              (categories[categoryId] or 0)+completedFlag;

            -- Initialise/update the category's parent categories
            _, parentId, _ = GetCategoryInfo(categoryId);
            while ((parentId ~= nil) and
                   (band((categories[parentId] or 0),
                     completedFlag) == 0)) do
              categories[parentId] = (categories[parentId] or 0)+completedFlag;
              _, parentId, _ = GetCategoryInfo(parentId);
            end
          end
        end
      end

      -- Update the categories
      AchievementFrameCategories_Update();
    end
  elseif (event == "ACHIEVEMENT_EARNED") then
    -- Update the categories and achievements, if required
    if (((AchievementFrameAchievements ~= nil) and
         (AchievementFrameAchievements:IsShown())) or
        ((AchievementFrameStats ~= nil) and
         (AchievementFrameStats:IsShown()))) then
      AchievementFrameCategories_Update();
    end
  end

  return
end

--
-- Function to clear all filtering of the achievements
--
achievementsFilterClear = function(update)
  -- Clear the achievement filtering
  if (nglAchievementsFrame ~= nil) then
    nglAchievementsFrame.searchBox:SetText("");
    nglAchievementsFrame.searchBox:ClearFocus();
    currentTabInfo.filtered.string = "";
  end

  -- Reset the Blizzard filtering to 'All'
  AchievementFrame_SetFilter(1);

  -- Force the achievements frame to update, if required
  if ((AchievementFrameAchievements:IsShown()) and (update)) then
    AchievementFrameAchievements_ForceUpdate();
  end

  return;
end

--
-- Function to enable/disable the achievement history buttons
--
achievementsHistoryBtnEnable = function()
  -- Check feature is enabled and the buttons exist
  if ((NigglesDBC.settings.UiAchievementSearch) and
      (nglAchievementsFrame ~= nil)) then
    -- Enable/disable the 'Back' and 'Forward' buttons
    nglAchievementsFrame.back:SetEnabled(
      currentTabInfo.history:IsMoveValid(true));
    nglAchievementsFrame.forward:SetEnabled(
      currentTabInfo.history:IsMoveValid(false));
  end

  return;
end

--
-- Function called when the tabs are updated
--
local function achievementsOnTabsUpdated()
  -- Check achievements aren't
  if (AchievementFrameComparison:IsShown()) then
    nglAchievementsFrame.searchBox:SetText("");
    nglAchievementsFrame.searchBox:Disable();
    nglAchievementsFrame.back:Disable();
    nglAchievementsFrame.forward:Disable();
  else
    -- Get info for the current tab
    currentTabInfo = tabInfo[(AchievementFrame.selectedTab or 1)];

    -- Set the search text
    nglAchievementsFrame.searchBox:SetText(
      currentTabInfo.filtered.string or "");
    nglAchievementsFrame.searchBox:Enable();

    -- Enable/disable the history buttons
    achievementsHistoryBtnEnable();
  end

  return;
end

--
-- Function called when the text changes in the Search box. If the change was
-- caused by the user the achievements are filtered.
--
local function achievementSearchOnTextChanged(self, userInput)
  -- Local Variables
  local searchText = self:GetText();
  local filteredInfo = currentTabInfo.filtered;

  -- Update the search box's appearance
  SearchBoxTemplate_OnTextChanged(self);

  -- Check if the search text has changed
  if (searchText ~= filteredInfo.string) then
    -- Check enough characters have been entered to make a search worthwhile
    if (strlen(searchText) >= MIN_CHARACTER_SEARCH) then
      -- Search the achievements for matches
      filteredInfo.string = searchText;
      SetAchievementSearchString(searchText);
    elseif (filteredInfo.string ~= "") then
      ClearAchievementSearchString();
      filteredInfo.categories   = {};
      filteredInfo.achievements = {};
      filteredInfo.string       = "";
      AchievementFrameCategories_Update();
      if ((AchievementFrame.selectedTab or 1) <= 2) then
        AchievementFrameAchievements_Update();
      else
        AchievementFrameStats_Update();
      end
    end
  end

  return;
end

--
-- Function called when a hyperlink is clicked in a chat frame. If the
-- hyperlink is for an achievement and it was right-clicked, the achievement
-- is displayed in the Achievement panel.
--
local function chatFrameOnHyperlinkShow(self, link, text, mouseButton)
  -- Local Variables
  local achievement = {};

  -- Check if the achievement has been right-clicked
  achievement.id = tonumber(string.match(text, "|Hachievement:(%d+):"));
  if ((NigglesDBC.settings.UiAchievementProgress == true) and
      (mouseButton == "RightButton") and
      (achievement.id ~= nil)) then
    -- Hide the Item Ref tooltip
    HideUIPanel(ItemRefTooltip);

    -- Load the Blizzard Achievement addon, if required;
    if (not AchievementFrame) then
      AchievementFrame_LoadUI();
    end

    -- Clear any filtering
    achievementsFilterClear(true);

    -- Display the achievement
    AchievementAlertFrame_OnClick(achievement);
  end

  return
end

--
-- Function hooked into 'SetItemRef' to change the title into a click-able
-- link
--
local function itemRefOnSet(link, text, mouseButton, chatFrame)
  -- Local Variables
  local showOverlay = false;

  -- Try to get the ID of the achievement being displayed by the tooltip
  itemRefAchievement.id   =
    tonumber(string.match(text, "|Hachievement:(%d+):"));

  -- Check an achievement is being displayed
  if (itemRefAchievement.id ~= nil) then
    -- Save required 'SetItemRef' info
    itemRefAchievement.link = link;
    itemRefAchievement.text = text;

    -- Check if the Achievement progress feature is currently enabled
    if (NigglesDBC.settings.UiAchievementProgress == true) then
      -- Update the tooltip's title
      ItemRefTooltipTextLeft1:SetText(text);
      ItemRefTooltip:Show();
      showOverlay = true;
    end
  end

  -- Show/hide the overlay that makes the tooltip's title click-able
  if (itemRefAchievement.overlay ~= nil) then
    itemRefAchievement.overlay:SetShown(showOverlay);
  end

  return;
end

--
-- Function called when the overlay over the Item Ref tooltip's first left text
-- font string is clicked. It displays the achievement currently being
-- displayed by the Item Ref tooltip in the Achievement panel.
--
local function itemRefOverlayOnClick(self, mouseButton)
  -- Load the Blizzard Achievement addon, if required;
  if (not AchievementFrame) then
    AchievementFrame_LoadUI();
  end

  -- Clear any filtering
  achievementsFilterClear(true);

  -- Display the achievement
  AchievementAlertFrame_OnClick(itemRefAchievement);

  return;
end

--
-- Function called 'PreClick' for a meta criteria button. It clears any filter
-- before the 'OnClick' script is called to make sure that the meta criteria
-- is visible.
--
metaCriteriaPreClick = function(self, button)
  -- Clear any filtering
  nglAchievementsFrame.searchBox:SetText("");
  nglAchievementsFrame.searchBox:ClearFocus();
  currentTabInfo.filtered.string = "";

  return;
end

--
-- Function called when the selected category, filter or search string changes.
-- It creates a list of achievements for the selected category based on the
-- filter and search string
--
selectedCategoryUpdate = function()
  -- Local Variables
  local achievementId;
  local band = bit.band;
  local categories = ACHIEVEMENTUI_CATEGORIES;
  local categoryId = currentTabInfo.blizzard.selectedCategory;
  local filterMask = 4-AchievementFrameFilterDropDown.value;
  local filterValue = AchievementFrameFilterDropDown.value;
  local filteredAchiev = currentTabInfo.filtered.achievements;
  local filteredCat = currentTabInfo.filtered.categories;
  local headerAdded = false;
  local numAchievements;
  local numComplete;
  local numIncomplete;
  local selectedAchiev = currentTabInfo.selected.achievements;
  local skipAchievement;

  -- Update the selected info for the category
  currentTabInfo.selected.categoryId  = categoryId;
  currentTabInfo.selected.filterValue = filterValue;
  currentTabInfo.selected.string      = currentTabInfo.filtered.string;
  wipe(currentTabInfo.selected.achievements);

  -- Update the info for the selected category
  if (((AchievementFrame.selectedTab or 1) <= 2) and
      (categoryId ~= "summary")) then
    -- Check there is a search string and the category has matches
    if ((currentTabInfo.filtered.string ~= "") and
        (band((filteredCat[categoryId] or 0), filterMask) ~= 0)) then
      -- Get the number of achievements in the category
      numAchievements, numComplete, numIncomplete =
        GetCategoryNumAchievements(categoryId);

      -- Work out which achievements should be shown
      for idx =
        (filterValue == 3 and numComplete+1 or 1),
        (filterValue == 2 and numComplete   or numAchievements) do
        -- Get the achievement's ID
        achievementId = GetAchievementInfo(categoryId, idx)

        -- Check if the achievement matches the search string
        if (band((filteredAchiev[achievementId] or 0), filterMask) ~= 0) then
          selectedAchiev[#selectedAchiev+1] = idx;
        end
      end
    end

    -- Update the scroll frame
    AchievementFrameAchievements_Update();
  elseif ((AchievementFrame.selectedTab or 1) == 3) then
    -- Check there is a search string and the category has matches
    if ((currentTabInfo.filtered.string ~= "") and
        (band((filteredCat[categoryId] or 0), filterMask) ~= 0)) then
      -- Add the category's header
      selectedAchiev[#selectedAchiev+1] = -categoryId;

      -- Work out which achievements should be shown for the category
      numAchievements = GetCategoryNumAchievements(categoryId);
      for idx = 1, numAchievements do
        _, skipAchievement, achievementId = GetStatistic(categoryId, idx);
        if ((not skipAchievement) and
            (band((filteredAchiev[achievementId] or 0), filterMask) ~= 0)) then
          selectedAchiev[#selectedAchiev+1] = achievementId;
        end
      end

      -- Work out which achievements should be shown from its sub-categories
      for idx, category in ipairs(categories) do
        -- Check the sub-category has achievements that should be shown
        if ((category.parent == categoryId) and
            (filteredCat[category.id] ~= nil)) then
          -- Add the category's header
          selectedAchiev[#selectedAchiev+1] = -category.id;

          -- Work out which achievements should be shown for the category
          numAchievements = GetCategoryNumAchievements(category.id);
          for idx = 1, numAchievements do
            _, skipAchievement, achievementId = GetStatistic(category.id, idx);
            if ((not skipAchievement) and
                (filteredAchiev[achievementId] ~= nil)) then
              selectedAchiev[#selectedAchiev+1] = achievementId;
            end
          end
        end
      end
    end

    -- Update the scroll frame
    AchievementFrameStats_Update();
  end

  return;
end

-------------------------------------------------------------------------------
--               G  L  O  B  A  L     F  U  N  C  T  I  O  N  S
-------------------------------------------------------------------------------

--
-- Function to enable/disable the Niggles Achievement functionality
--
function NigglesAchievementsEnable()
  -- Local Variables
  local frame;
  local fontString;

  -- Check if the Niggles Achievement frame should be created
  if ((AchievementFrame ~= nil) and
      (NigglesDBC.settings.UiAchievementSearch) and
      (nglAchievementsFrame == nil)) then
    -- Create the Niggles Achievement frame
    frame = CreateFrame("FRAME", nil, AchievementFrame,
      "NigglesAchievementsTemplate");
    frame:SetBackdropBorderColor(
      ACHIEVEMENTUI_GOLDBORDER_R, ACHIEVEMENTUI_GOLDBORDER_G,
      ACHIEVEMENTUI_GOLDBORDER_B, ACHIEVEMENTUI_GOLDBORDER_A);
    frame:SetScript("OnEvent", achievementSearchOnEvent);
    frame:SetScript("OnShow",  achievementSearchOnShow);
    frame.searchBox:SetScript("OnTextChanged", achievementSearchOnTextChanged);
    frame.back:SetScript("OnClick", achievementHistoryBtnOnClick);
    frame.forward:SetScript("OnClick", achievementHistoryBtnOnClick);
    nglAchievementsFrame = frame;

    -- Create a new edge to replace the existing backdrop
    frame = CreateFrame("FRAME", nil, AchievementFrameCategories);
    frame:SetPoint("TOPLEFT", 0, -37);
    frame:SetPoint("BOTTOMRIGHT", 0, 0);
    frame:SetBackdrop(AchievementFrameCategories:GetBackdrop());
    frame:SetBackdropBorderColor(ACHIEVEMENTUI_GOLDBORDER_R,
      ACHIEVEMENTUI_GOLDBORDER_G, ACHIEVEMENTUI_GOLDBORDER_B,
      ACHIEVEMENTUI_GOLDBORDER_A);
    AchievementFrameCategories:SetBackdrop(nil);
    AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -42);
    nglAchievementsEdge = frame;

    -- Create a font string to display messages for the Stats tab
    fontString = AchievementFrameStatsContainer:CreateFontString(nil,
      "ARTWORK", "GameFontHighlight");
    fontString:SetSize(490, 0);
    fontString:SetPoint("TOP",
      AchievementFrameAchievementsFeatOfStrengthText, 0, 0);
    fontString:SetText(L["No Matches"]);
    nglAchievementsMsg = fontString;

    -- Hook Blizzard functions that require action
    hooksecurefunc("AchievementFrame_UpdateTabs", achievementsOnTabsUpdated);
    hooksecurefunc("AchievementFrame_SetFilter", achievementsFilterOnSet);
    hooksecurefunc("AchievementFrameCategories_Update",
      achievementCategoriesUpdate);
    hooksecurefunc("AchievementButton_GetMeta", achievementMetaBtnCreated);

    -- Save the Blizzard functions used to update displaying of achievements
    blizzardAchievementsUpdate = AchievementFrameAchievements_Update;
    blizzardStatisticsUpdate   = AchievementFrameStats_Update;

    -- Get the Blizzard info for the tabs
    tabInfo[1].blizzard = ACHIEVEMENT_FUNCTIONS;
    tabInfo[2].blizzard = GUILD_ACHIEVEMENT_FUNCTIONS;
    tabInfo[3].blizzard = STAT_FUNCTIONS;
  end

  -- Check the Niggles Achievement frame exists
  if (nglAchievementsFrame ~= nil) then
    -- Show/hide the Niggles Achievement frame
    if (NigglesDBC.settings.UiAchievementSearch == true) then
      -- Hide the Blizzard search box frames/textures
      AchievementFrame.searchBox:Hide();
      AchievementFrameHeaderRightDDLInset:Hide();
      AchievementFrame_HideSearchPreview();

      -- Get the info for the current tab
      currentTabInfo = tabInfo[(AchievementFrame.selectedTab or 1)];
      if (AchievementFrameCategoriesContainer.buttons ~= nil) then
        AchievementFrameCategories_Update();
      end

      -- Show the Niggles achievements frames
      nglAchievementsFrame:Show();
      nglAchievementsEdge:SetPoint("TOPLEFT", 0, -37);
      AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -42);

      -- Hide the 'No Matches' message
      nglAchievementsMsg:Hide();

      -- Register for events
      nglAchievementsFrame:RegisterEvent("ACHIEVEMENT_SEARCH_UPDATED");
      nglAchievementsFrame:RegisterEvent("ACHIEVEMENT_EARNED");

      -- Override the Blizzard functions used to update the lists
      AchievementFrameAchievements_Update = achievementsListUpdate;
      AchievementFrameStats_Update        = achievementsStatsUpdate;
    else
      -- Clear the filtered and selected info
      for _, current in ipairs(tabInfo) do
        current.filtered.string = "";
        wipe(current.filtered.categories);
        wipe(current.filtered.achievements);
        current.selected.string      = "";
        current.selected.categoryId  = -1;
        current.selected.filterValue = 0;
        wipe(current.selected.achievements);
      end
      nglAchievementsFrame.searchBox:SetText("");

      -- Hide the Niggles achievements elements
      nglAchievementsFrame:Hide();
      nglAchievementsEdge:SetPoint("TOPLEFT", 0, 0);
      AchievementFrameCategoriesContainer:SetPoint("TOPLEFT", 0, -5);
      AchievementFrameCategories_Update();
      nglAchievementsMsg:Hide();

      -- Unregister for events
      nglAchievementsFrame:UnregisterEvent("ACHIEVEMENT_SEARCH_UPDATED");
      nglAchievementsFrame:UnregisterEvent("ACHIEVEMENT_EARNED");

      -- Show the Blizzard search box frames/textures
      AchievementFrame.searchBox:Show();
      AchievementFrameHeaderRightDDLInset:Show();
      AchievementFrame_HideSearchPreview();

      -- Restore the Blizzard functions used to update the lists
      AchievementFrameAchievements_Update = blizzardAchievementsUpdate;
      AchievementFrameStats_Update        = blizzardStatisticsUpdate;
    end

    -- Update some Blizzard tables with the current functions
    ACHIEVEMENT_FUNCTIONS.updateFunc =
      AchievementFrameAchievements_Update;
    GUILD_ACHIEVEMENT_FUNCTIONS.updateFunc =
      AchievementFrameAchievements_Update;
    AchievementFrameAchievementsContainer.update =
      AchievementFrameAchievements_Update;
    STAT_FUNCTIONS.updateFunc =
      AchievementFrameStats_Update;
    AchievementFrameStatsContainer.update =
      AchievementFrameStats_Update;

    -- Force the categories and achievements to be updated
    if (AchievementFrameCategoriesContainer:IsVisible()) then
      AchievementFrameCategories_Update();
    end
    if (AchievementFrameAchievementsContainer:IsVisible()) then
      AchievementFrameAchievements_Update();
    end
    if (AchievementFrameStatsContainer:IsVisible()) then
      AchievementFrameStats_Update();
    end
  end

  -- Check if the Achievement progress feature is enabled
  if (NigglesDBC.settings.UiAchievementProgress) then
    -- Create the overlay frame used to make the tooltip title click-able.
    if (itemRefAchievement.overlay == nil) then
      frame = CreateFrame("FRAME", nil, ItemRefTooltip);
      frame:SetAllPoints(ItemRefTooltipTextLeft1);
      frame:EnableMouse();
      frame:SetScript("OnMouseUp", itemRefOverlayOnClick);
      itemRefAchievement.overlay = frame;
    end

    -- Update the Item Ref tooltip, if currently displaying an achievement
    if ((ItemRefTooltip:IsShown()) and
        (itemRefAchievement.id ~= nil)) then
    end
  end

  -- Update the Item Ref tooltip, if currently displaying an achievement
  if ((ItemRefTooltip:IsShown()) and
      (itemRefAchievement.id ~= nil)) then
    HideUIPanel(ItemRefTooltip);
    SetItemRef(itemRefAchievement.link, itemRefAchievement.text,
     "LeftButton", nil);
  end

  return;
end

--
-- Function to initialise the Achievements functionality
--
function NigglesAchievementsInitialise()
  -- Hook the functions for displaying achievements
  hooksecurefunc("ChatFrame_OnHyperlinkShow", chatFrameOnHyperlinkShow);
  hooksecurefunc("SetItemRef", itemRefOnSet);

  -- Hook into the creation of the Achievement Alert frames
  hooksecurefunc("AchievementAlertFrame_OnClick", achievementAlertOnClick);

  return;
end

--
-- Function called when Blizzard's Achievements addon is loaded.
--
function NigglesAchievementsOnLoad()
  -- Hook the functions used to select achievements
  hooksecurefunc("AchievementFrame_SelectAchievement",
    achievementOnSelect);
  hooksecurefunc("AchievementFrameAchievements_SelectButton",
    achievementOnSelect);

  -- Enable/disable the Niggles Achievements features
  NigglesAchievementsEnable();

  return;
end
